from climb.exceptions import UnknownCommand


def command(function):
    function.command = True
    return function


def completers(*compl):
    def wrapper(function):
        function.completers = compl
        return function
    return wrapper


class Commands(object):

    def __init__(self, cli):
        self._cli = cli
        self._commands = {}

    def execute(self, name, *args, **kwargs):
        if hasattr(self, name):
            method = getattr(self, name)
            if getattr(method, 'command', None):
                return method(*args, **kwargs)

        raise UnknownCommand("There is no action for command {}".format(command))

    def get_completer(self, name, position):
        if hasattr(self, name):
            method = getattr(self, name)
            compl = getattr(method, 'completers', [])

            if compl:
                pos = position-1
                if len(compl) > pos:
                    return compl[pos]
                else:
                    return compl[0]

        raise UnknownCommand("No completer for command {}".format(command))

    @command
    def help(self, parser, all_commands, subject):
        if subject:
            subparsers = [cmd for cmd in all_commands
                          if cmd.name == subject]
            if subparsers:
                parser = subparsers[0].parser

        return parser.print_help()

    @command
    def exit(self):
        self._cli.set_running(False)



import os
import shlex
import readline

from climb.config import load_config, config
from climb.exceptions import CLIException
from climb.paths import ROOT_PATH


class Climb(object):
    _prompt = "[{path}]> "
    _running = False

    def __init__(self, name, args, commands, completer, prompt=None, skip_delims=None):
        self._name = name
        load_config(name)

        self._current_path = ROOT_PATH
        self._verbose = config.getboolean(name, 'verbose')
        self._history_file = os.path.expanduser(config[name].get('history', ''))

        self._args = args(self)
        self._commands = commands(self)
        self._completer = completer(self)

        if prompt is not None:
            self._prompt = prompt

        self._skip_delims = skip_delims

    def run(self):
        """Loops and executes commands in interactive mode."""
        if self._skip_delims:
            delims = readline.get_completer_delims()
            for delim in self._skip_delims:
                delims = delims.replace(delim, '')
            readline.set_completer_delims(delims)

        readline.parse_and_bind("tab: complete")
        readline.set_completer(self._completer.complete)

        if self._history_file:
            # Ensure history file exists
            if not os.path.isfile(self._history_file):
                open(self._history_file, 'w').close()

            readline.read_history_file(self._history_file)

        self._running = True
        try:
            while self._running:
                try:
                    command = input(self._format_prompt())
                    if command:
                        result = self.execute(*shlex.split(command))
                        if result:
                            print(result)
                except CLIException as exc:
                    print(exc)
                except (KeyboardInterrupt, EOFError):
                    self._running = False
                    print()
        finally:
            if self._history_file:
                readline.write_history_file(self._history_file)

    def _format_prompt(self):
        return self._prompt.format(path=self._current_path)

    def execute(self, *args):
        """Executes single command and returns result."""
        command, kwargs = self.parse(*args)
        return self._commands.execute(command, **kwargs)

    def parse(self, *args):
        parsed = self._args.parse(*args)

        kwargs = dict(parsed._get_kwargs())
        command = kwargs.pop('command')

        return command, kwargs

    def log(self, message, *args, **kwargs):
        if self._verbose:
            print(message.format(*args, **kwargs))

    def set_running(self, running):
        self._running = running

    def set_current_path(self, current_path):
        self._current_path = current_path

    @property
    def args(self):
        return self._args

    @property
    def commands(self):
        return self._commands

    @property
    def current_path(self):
        return self._current_path



ROOT_PATH = "/"
SEPARATOR = "/"


def format_path(current_path, path, default=None):
    if path:
        path = path.strip()

    if not path:
        return default if default else current_path

    absolute = path.startswith(ROOT_PATH)
    if not absolute:
        path = SEPARATOR.join((current_path, path))

    parts = split_path(path)

    result = []
    for part in parts:
        if part == ".":
            continue
        elif part == "..":
            # Up one level
            result = result[:-1]
        else:
            result.append(part)

    new_path = ROOT_PATH + SEPARATOR.join(result)

    return new_path


def split_path(path):
    if not path:
        return []

    return [part for part in path.strip().split(SEPARATOR)
            if part]


import os
import glob
import readline


def current_argument(buffer, begin, end):
    trimmed_buffer = buffer[:end]
    position = ' '.join(trimmed_buffer.split()).count(' ')
    if trimmed_buffer.endswith(' '):
        position += 1

    space_position = buffer.rfind(' ', 0, begin)
    argument = buffer[space_position+1:end]

    return position, argument


class Completer(object):

    def __init__(self, cli):
        self._cli = cli
        self._commands = [command.name for command in self._cli.args.commands]

    def complete(self, text, state):
        buffer = readline.get_line_buffer()
        begin = readline.get_begidx()
        end = readline.get_endidx()

        position, argument = current_argument(buffer, begin, end)
        if position > 0:
            command, _ = self._cli.parse(*buffer.split())
            completer_name = self._cli.commands.get_completer(command, position)
            completer = getattr(self, completer_name)
        else:
            completer = self.commands

        completions = completer(argument, text)

        if state < len(completions):
            return completions[state]
        else:
            return None

    def commands(self, arg, text):
        commands = [c for c in self._commands
                    if c.startswith(arg)]

        if len(commands) == 1:
            return ["{} ".format(commands[0])]

        return commands

    def system_path(self, arg, text):
        glob_path = "{}*".format(os.path.expanduser(arg))

        paths = [path for path in glob.iglob(glob_path)
                 if path.startswith(arg)]

        if len(paths) == 1:
            child = paths[0].split('/')[-1]
            if os.path.isdir(paths[0]):
                return ["{}/".format(child)]
            else:
                return ["{} ".format(child)]

        return [path.split('/')[-1] for path in paths]



import os
import configparser

from climb.exceptions import ConfigNotFound

DEFAULT_CONFIG_PATHS = [
    './{name}.conf',
    '~/.{name}.conf',
    '/etc/{name}/{name}.conf',
]

config = configparser.ConfigParser()


def load_config(name):
    config.clear()

    paths = [path.format(name=name) for path in DEFAULT_CONFIG_PATHS]
    for config_path in paths:
        if _read_config(config_path):
            break
    else:
        raise ConfigNotFound("Could not find {name}.conf".format(name))


def load_config_file(path):
    config.clear()

    if not _read_config(path):
        raise ConfigNotFound("Could not load {}".format(path))


def _read_config(path):
    config_path = os.path.expanduser(path)
    if os.path.isfile(config_path) and os.access(config_path, os.R_OK):
        config.read(config_path)
        return True
    else:
        return False



#!/usr/bin/python3
from setuptools import setup, find_packages

setup(name='climb',
      version='0.3.1',
      description='Library for interactive command line applications.',
      author='Milosz Smolka',
      author_email='m110@m110.pl',
      url='https://github.com/m110/climb',
      packages=find_packages(exclude=['tests']),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: MIT License',
          'Operating System :: POSIX',
          'Programming Language :: Python :: 3.4',
          'Topic :: System :: Systems Administration',
      ])



import argparse
from collections import namedtuple

from climb.exceptions import UnknownCommand

Command = namedtuple("Command", ['name', 'parser'])


class ArgsParser(argparse.ArgumentParser):

    def error(self, message):
        raise UnknownCommand(self.format_help().strip())


class Args(object):

    def __init__(self, cli):
        self._cli = cli
        self._commands = []

        self._parser = ArgsParser(add_help=False)
        self._commands_parser = self._parser.add_subparsers(help="command")

        self._load_commands()

        help = self._add_command("help", "show detailed help for command",
                                 parser=self._parser, all_commands=self._commands)
        help.add_argument("subject", nargs="?", default=None)

        self._add_command("exit", "exit console")

        # Store all subparsers for improved help messages and completion support
        actions = [action for action in self._parser._actions
                   if isinstance(action, argparse._SubParsersAction)]
        self._commands.extend([Command(choice, subparser)
                               for action in actions
                               for choice, subparser in action.choices.items()])

    def _load_commands(self):
        """Should be overridden by subclass."""
        pass

    def _add_command(self, name, help, **kwargs):
        command = self._commands_parser.add_parser(name, help=help, add_help=False)
        command.set_defaults(command=name, **kwargs)
        return command

    def parse(self, *args):
        return self._parser.parse_args(args)

    @property
    def commands(self):
        return self._commands
